/*
   sptrace.c - Secure ptrace() module

   Copyright (C) 2003, 2004 Krzysztof Burghardt <einstein@underground.org.pl>.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
   Version Information
*/
#define VERSION "1.4.0"
#define AUTHOR  "Krzysztof Burghardt <einstein@underground.org.pl>"
#define DESC    "Secure ptrace() Linux Kernel Module (LKM)"
#define LICENSE "GPL"

#ifndef __KERNEL__
# define __KERNEL__
#endif

#ifndef MODULE
# define MODULE
#endif

/*
   Includes
*/
#include <linux/version.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
# include <linux/modversions.h>
#endif
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/config.h>
#include <linux/sched.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
# include <sys/syscall.h>
#else
# include <linux/unistd.h>
#endif 

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 // Please remove or comment out this line:
# error You did not RTFM carefully enough !!!
 // Then run: 
 //          grep sys_call_table /boot/System.map
 // or:
 //          grep sys_call_table /proc/kallsyms
 // and fill in sys_call_table address in line below...
# define SYS_CALL_TABLE_ADDR 0x????????
 // Linus, why you removed sys_call_table? It is not more secure anyway!
 void **sys_call_table;
#else
 extern void **sys_call_table;
#endif

/*
   Global Variables
*/

/* GID of group allowed to use ptrace() call */
static int ptrace_group = 0;

/* Pointer to original ptrace call */
static int (*orig_ptrace) (int, int, int, int);

/*
   Helper Function
*/
static inline void
alert(const task_t * who, const char *message)
{
    static int last_pid = 0;

    if (last_pid != who->pid) {
	last_pid = who->pid;

	printk("sptrace: %s (%s:%d) UID(%d) EUID(%d), "
	       "parent (%s:%d) UID(%d) EUID(%d)\n", message,
#	if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	     who->comm, who->pid, who->uid, who->euid,
	     who->parent->comm, who->parent->pid,
	     who->parent->uid, who->parent->euid);
#	else
	     who->comm, who->pid, who->uid, who->euid,
	     who->p_pptr->comm, who->p_pptr->pid, who->p_pptr->uid,
	     who->p_pptr->euid);
#	endif
    }

    return;

}

/*
   Security
*/
static int
sptrace(int request, int pid, int addr, int data)
{
	if (current->euid == 0 || in_group_p(ptrace_group) > 0) {
		alert(current, "ptrace() by");
		return (orig_ptrace) (request, pid, addr, data);
	}

	alert(current, "ptrace() DENIED for");

	return -EPERM;
}

/*
   init function
*/
static int __init
sptrace_init_module(void)
{
#	if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	sys_call_table = (void **) SYS_CALL_TABLE_ADDR;
#	endif
	printk("sptrace: secure ptrace() module v%s (ptrace group: %d)\n",
		VERSION, ptrace_group);
	orig_ptrace = sys_call_table[__NR_ptrace];
	sys_call_table[__NR_ptrace] = sptrace;
	return 0;
}

/*
   exit function
*/
static void __exit
sptrace_cleanup_module(void)
{
	printk("sptrace: unloading secure ptrace() module\n");
	sys_call_table[__NR_ptrace] = orig_ptrace;
	return;
}

/*
   module_init & module_exit
*/
module_init(sptrace_init_module);
module_exit(sptrace_cleanup_module);

/*
   Author and description
*/

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESC);
MODULE_LICENSE(LICENSE);

/*
   Parameters
*/

MODULE_PARM(ptrace_group, "i");
MODULE_PARM_DESC(ptrace_group,
		 "GID of group allowed to ptrace (default: 0 - no ptrace group)");

/*
   EOF
*/
